/*
 * Copyright 2018 by Kappich Systemberatung Aachen
 * ALL RIGHTS RESERVED.
 */

package de.bsvrz.dav.daf.accessControl;

import de.bsvrz.dav.daf.main.*;
import de.bsvrz.dav.daf.main.config.*;
import de.bsvrz.sys.funclib.debug.Debug;
import de.bsvrz.sys.funclib.kappich.annotations.Nullable;

/**
 * Empfängt den Status der Rechteprüfung über die Attributgruppe
 * atg.datenverteilerRechteprüfung am Datenverteilerobjekt. Falls das Datenmodell zu alt ist,
 * tut diese Klasse nichts.
 *
 * @author Kappich Systemberatung
 * @version $Revision$
 */
public class DafAccessControlReceiver {
	private static final Debug _debug = Debug.getLogger();
	private final ClientDavInterface _connection;
	private final ConfigurationObject _davObject;
	private final DataModel _dataModel;
	private DataDescription _dataDescription = null;
	private static final int TIMEOUT_SECONDS = 10;

	/**
	 * Erstellt einen neuen DafAccessControlReceiver, der den angeschlossenen Datenverteiler nach der Art der Rechteprüfung fragt  
	 * @param connection Verbindung
	 * 
	 */
	public DafAccessControlReceiver(final ClientDavInterface connection) {
		this(connection, connection.getLocalDav());
	}

	/**
	 * Erstellt einen neuen DafAccessControlReceiver, der einen beliebigen Datenverteiler nach der Art der Rechteprüfung fragt
	 * @param connection Verbindung
	 * @param davObject Datenverteiler-Objekt
	 */
	public DafAccessControlReceiver(final ClientDavInterface connection, final ConfigurationObject davObject) {
		_connection = connection;
		_davObject = davObject;
		_dataModel = _connection.getDataModel();
		AttributeGroup atg = _dataModel.getAttributeGroup("atg.datenverteilerRechteprüfung");
		Aspect asp = _dataModel.getAspect("asp.standard");
		if(atg == null || asp == null) return;
		_dataDescription = new DataDescription(atg, asp);
	}

	/**
	 * Gibt die Art der Rechteprüfung zurück
	 * @return Art der Rechteprüfung oder null, falls diese nicht bestimmt werden konnte (z. B. zu alte Kernsoftware)
	 */
	@Nullable
	public AccessControlMode getAccessControlMode() {
		if(_dataDescription == null) {
			_debug.warning("Art der Rechteprüfung konnte nicht bestimmt werden, Datenmodell veraltet.");
			return null;
		}
		int i = 0;
		ResultData data = null;
		while(i < TIMEOUT_SECONDS) {
			data = _connection.getData(_davObject, _dataDescription, 10000);
			if(data.hasData()) {
				return extractMode(data.getData());
			}
			i++;
		}
		_debug.warning("Art der Rechteprüfung konnte nicht bestimmt werden, timeout beim Warten auf Daten: " + data);
		return null;
	}

	@Nullable
	private static AccessControlMode extractMode(final Data data) {
		switch(data.getTextValue("Rechteprüfung").getText()) {
			case "Alte Rechteprüfung": return AccessControlMode.OldDataModel;
			case "Neue Rechteprüfung": return AccessControlMode.NewDataModel;
			case "Deaktiviert": return AccessControlMode.Disabled;
			default: {
				_debug.warning("Art der Rechteprüfung konnte nicht bestimmt werden, unbekannter Wert: " + data);
				return null;
			}
		}
	}

}
