/*
 * Copyright 2004 by Kappich+Kniß Systemberatung Aachen (K2S)
 * Copyright 2007-2018 by Kappich Systemberatung, Aachen
 *
 * This file is part of de.bsvrz.pat.sysbed.
 *
 * de.bsvrz.pat.sysbed is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * de.bsvrz.pat.sysbed is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with de.bsvrz.pat.sysbed.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Contact Information:
 * Kappich Systemberatung
 * Martin-Luther-Straße 14
 * 52062 Aachen, Germany
 * phone: +49 241 4090 436
 * mail: <info@kappich.de>
 */

package de.bsvrz.pat.sysbed.preselection.treeFilter.standard;

import de.bsvrz.dav.daf.main.ClientDavInterface;
import de.bsvrz.dav.daf.main.config.*;
import de.bsvrz.pat.sysbed.preselection.treeFilter.plugins.api.ExtendedFilter;
import de.bsvrz.sys.funclib.configObjectAcquisition.ConfigurationHelper;
import de.bsvrz.sys.funclib.debug.Debug;

import java.util.*;

/**
 * Die Klasse {@code Filter} speichert ein Kriterium und die dazugehörigen Werte und bietet Methoden an, um Systemobjekte anhand des Kriteriums zu
 * filtern.
 *
 * @author Kappich Systemberatung
 * @version $Revision$
 */
public class Filter {

	/** Der Debug-Logger der Klasse */
	private static Debug _debug = Debug.getLogger();

	/** speichert das Kriterium des Filters */
	private String _criteria;

	/** speichert die Einträge zum Kriterium */
	private String[] _values;

	/** speichert die Verbindung zum Datenverteiler */
	private ClientDavInterface _connection;

	/** filtert die aufgelisteten Objekte nach dem Konfigurationsbereich */
	public static final String CONFIGURATIONAREA = "Konfigurationsbereich";

	/** filtert die aufgelisteten Objekte nach dem Objekttyp */
	public static final String OBJECTTYPE = "Objekttyp";

	/** filtert die aufgelisteten Objekte nach der Attributgruppe */
	public static final String ATTRIBUTEGROUP = "Attributgruppe";

	/** filtert die aufgelisteten Objekte nach dem Aspekt */
	public static final String ASPECT = "Aspekt";

	/** filtert die aufgelisteten Objekte nach dem Objekt */
	public static final String OBJECT = "Objekt";

	/** filtert die aufgelisteten Objekte nach einem erweiterten Filter */
	public static final String EXTENDED = "Erweitert";


	/**
	 * Erzeugt ein neues Objekt der Klasse {@code Filter}.
	 *
	 * @param criteria   Filterkriterium
	 * @param values     Filterattribute
	 * @param connection Verbindung zum Datenverteiler
	 */
	public Filter(String criteria, String[] values, ClientDavInterface connection) {
		_criteria = criteria;
		_values = values;
		_connection = connection;
	}

	/**
	 * Die übergebenen Systemobjekte werden entsprechend des Kriteriums gefiltert und zurückgegeben.
	 *
	 * @param systemObjects die zu filternden Systemobjekte
	 *
	 * @return die gefilterten Systemobjekte
	 */
	public Collection<SystemObject> filterObjects(Collection<SystemObject> systemObjects) {
		Collection<SystemObject> tempObjects;
		if(systemObjects == null) {
			tempObjects = new LinkedList<>();
		}
		else {
			tempObjects = applyFilter(systemObjects);
		}
		return Collections.unmodifiableCollection(tempObjects);
	}

	/**
	 * Die Systemobjekte werden in Abhängigkeit vom Kriterium gefiltert.
	 *
	 * @param systemObjects die zu filternden System-Objekte
	 *
	 * @return die gefilterten System-Objekte
	 */
	private Collection<SystemObject> applyFilter(Collection<SystemObject> systemObjects) {
		DataModel configuration = _connection.getDataModel();
		Collection<SystemObject> tempObjects = new HashSet<>();
		if(_criteria.equals("Konfigurationsbereich")) {
			for(String value : _values) {
				List resultList;
				try {
					resultList = ConfigurationHelper.getObjects(value, configuration);
				}
				catch(IllegalArgumentException e) {
					_debug.warning("Zum Wert " + value + " kein passendes Objekt gefunden!");
					resultList = new ArrayList();
				}
				if(resultList.size() > 0) {
					if(resultList.size() > 1) _debug.finest("Anzahl der erstellten ObjekttypenObjekte: " + resultList.size());
					for(final Object aResultList : resultList) {
						ConfigurationArea configurationArea = (ConfigurationArea) aResultList;
						for(SystemObject systemObject : systemObjects) {
							ConfigurationArea confArea = systemObject.getConfigurationArea();
							if(confArea == configurationArea) {
								tempObjects.add(systemObject);
							}
						}
					}
				}
				else {
					_debug.warning("Zum Wert " + value + " kein passendes Objekt gefunden!");
				}
			}
		}
		else if(_criteria.equals("Objekttyp")) {
			for(String value : _values) {
				List resultList;
				try {
					resultList = ConfigurationHelper.getObjects(value, configuration);
				}
				catch(IllegalArgumentException e) {
					_debug.warning("Zum Wert " + value + " kein passendes Objekt gefunden!");
					resultList = new ArrayList();
				}
				if(resultList.size() > 0) {
					if(resultList.size() > 1) _debug.finest("Anzahl der erstellten ObjekttypenObjekte: " + resultList.size());
					// Typen und Subtypen sammeln, isOfType is zu langsam für jedes einzelne Objekt
					final Set<SystemObjectType> relevantTypes = new LinkedHashSet<>();
					for(final Object aResultList : resultList) {
						SystemObjectType systemObjectType = (SystemObjectType) aResultList;
						addSubTypes(relevantTypes, systemObjectType);
					}
					for(SystemObject systemObject : systemObjects) {
						if(relevantTypes.contains(systemObject.getType())) {
							tempObjects.add(systemObject);
						}
					}
				}
				else {
					_debug.warning("Zum Wert " + value + " kein passendes Objekt gefunden!");
				}
			}
		}
		else if(_criteria.equals("Attributgruppe")) {
			// AttributGruppen für alle Values holen
			for(String value : _values) {
				List resultList;
				try {
					resultList = ConfigurationHelper.getObjects(value, configuration);
				}
				catch(IllegalArgumentException e) {
					_debug.warning("Zum Wert " + value + " kein passendes Objekt gefunden!");
					resultList = new ArrayList();
				}
				if(resultList.size() > 0) {
					if(resultList.size() > 1) _debug.finest("Anzahl der erstellten AttributgruppenObjekte: " + resultList.size());
					for(final Object aResultList : resultList) {
						AttributeGroup attributeGroup = (AttributeGroup) aResultList;
						for(SystemObject systemObject : systemObjects) {
							List atgs = systemObject.getType().getAttributeGroups();
							if(atgs.contains(attributeGroup)) {
								tempObjects.add(systemObject);
							}
						}
					}
				}
				else {
					_debug.warning("Zum Wert " + value + " kein passendes Objekt gefunden!");
				}
			}
		}
		else if(_criteria.equals("Aspekt")) {
			for(String value : _values) {
				List resultList;
				try {
					resultList = ConfigurationHelper.getObjects(value, configuration);
				}
				catch(IllegalArgumentException e) {
					_debug.warning("Zum Wert " + value + " kein passendes Objekt gefunden!");
					resultList = new ArrayList();
				}
				if(resultList.size() > 0) {
					if(resultList.size() > 1) _debug.finest("Anzahl der erstellten AspektObjekte: " + resultList.size());
					for(final Object aResultList : resultList) {
						Aspect aspect = (Aspect) aResultList;
						for(SystemObject systemObject : systemObjects) {
							List atgs = systemObject.getType().getAttributeGroups();
							for(final Object atg : atgs) {
								AttributeGroup attributeGroup = (AttributeGroup) atg;
								Collection aspects = attributeGroup.getAspects();
								if(aspects.contains(aspect)) {
									tempObjects.add(systemObject);
								}
							}
						}
					}
				}
				else {
					_debug.warning("Zum Wert " + value + " kein passendes Objekt gefunden!");
				}
			}
		}
		else if(_criteria.equals("Objekt")) {
			// Objekt holen zur Pid und dann alle durchlaufen und vergleichen
			for(String value : _values) {
				List resultList;
				try {
					resultList = ConfigurationHelper.getObjects(value, configuration);
				}
				catch(IllegalArgumentException e) {
					_debug.warning("Zum Wert " + value + " kein passendes Objekt gefunden!");
					resultList = new ArrayList();
				}
				if(resultList.size() > 0) {
					if(resultList.size() > 1) _debug.finest("Anzahl der erstellten Objekte: " + resultList.size());
					for(final Object aResultList : resultList) {
						SystemObject systemObject = (SystemObject) aResultList;
						if(systemObjects.contains(systemObject)) {
							tempObjects.add(systemObject);
						}
					}
				}
				else {
					_debug.warning("Zum Wert " + value + " kein passendes Objekt gefunden!");
				}
			}
		}
		else if(_criteria.equals("Erweitert")) {
			try {
				// Klassenname mit Pfad angeben z.B. sys.funclib.preselection.test.Filtertest
				String str = _values[0];
				Class c = Class.forName(str);
				ExtendedFilter extendedFilter = (ExtendedFilter)c.newInstance();
				extendedFilter.setValues(_values);
				extendedFilter.setConnection(_connection);
				tempObjects = extendedFilter.applyFilter(systemObjects);
			}
			catch(ClassNotFoundException e) {
				e.printStackTrace();
			}
			catch(IllegalAccessException e) {
				e.printStackTrace();
			}
			catch(InstantiationException e) {
				e.printStackTrace();
			}
		}
		else {
			_debug.error("Dieses Kriterium '" + _criteria + "' wird nicht unterstützt. Verwenden Sie dafür einen Erweiterten Filter.");
			tempObjects = systemObjects;
		}
		return tempObjects;
	}

	private void addSubTypes(final Set<SystemObjectType> relevantTypes, final SystemObjectType systemObjectType) {
		if(relevantTypes.contains(systemObjectType)) return;
		List<SystemObjectType> subTypes = systemObjectType.getSubTypes();
		for(SystemObjectType subType : subTypes) {
			addSubTypes(relevantTypes, subType);
		}
		relevantTypes.add(systemObjectType);
	}
}
